(***********************************************************************)
(*                                                                     *)
(*                             Pio (OCaml)                             *)
(*                                                                     *)
(*                    Polyglot Input/Output Library                    *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2005-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: read_typed_value.ml,v 1.13 2014/01/10 19:22:10 fclement Exp $ *)

open Value_type_def;;
open Layout;;
open Typed_value;;

let debug fun_name =
  Options_pio.debug (Printf.sprintf "Read_typed_value.%s" fun_name)
;;

let fatal_error fun_name msg =
  Options_pio.fatal_error
    (Printf.sprintf "Read_typed_value.%s" fun_name) msg 2
;;

(* Read the type of a typed value *)
let r_value_type ib =
  assert (debug "r_value_type" "start reading");
  Scanf.bscanf ib " " ();
  let t = Value_type.of_ib ib in
  assert (debug "r_value_type" "finish reading");
  t
;;

(* Read the layout of a multi-dimensional array *)
let r_layout ib =
  let l = Layout.of_ib ib in
  Scanf.bscanf ib "\n" l
;;

let r_size ib = Scanf.bscanf ib "<%i>\n" (fun n -> n);;
let r_size_matrix ib = Scanf.bscanf ib "<%i, %i>\n" (fun n1 n2 -> n1, n2);;

(* r_tuple *)
let r_couple r_c ib =
  assert (debug "r_couple" "start reading");
  Scanf.bscanf ib "(" ();
  let x = r_c ib in
  Scanf.bscanf ib ", " ();
  let y = r_c ib in
  Scanf.bscanf ib ")" ();
  assert (debug "r_couple" "finish reading");
  x, y

and r_triple r_c ib =
  assert (debug "r_triple" "start reading");
  Scanf.bscanf ib "(" ();
  let x = r_c ib in
  Scanf.bscanf ib ", " ();
  let y = r_c ib in
  Scanf.bscanf ib ", " ();
  let z = r_c ib in
  Scanf.bscanf ib ")" ();
  assert (debug "r_triple" "finish reading");
  x, y, z
;;

let read_8_bytes ib =
  Scanf.bscanf ib "&<%i>" (fun i ->
    if i <> 8 then
      failwith
        "read_8_bytes: \
         binary representation for float should be 8-byte long." else
    Scanf.bscanf ib "%8[\000-\255]" (fun s -> s))
;;
(* My_string.init 8 (fun i -> Scanf.bscanf ib "%c" (fun c -> c));; *)

(* r_simple_lexem *)
let r_bool ib =
  assert (debug "r_bool" "start reading");
  let b = Scanf.bscanf ib "%B" (fun b -> b) in
  assert (debug "r_bool" "finish reading");
  b

and r_string ib =
  assert (debug "r_string" "start reading");
  let n = r_size ib in
  Scanf.bscanf ib "%S"
    (fun s ->
      let l = String.length s in
      if l = n then (assert (debug "r_string" "finish reading"); s) else
      failwith
        (Printf.sprintf
           "String %S has length %i but should have length %i!" s l n))

and r_int ib =
  assert (debug "r_int" "start reading");
  let i = Scanf.bscanf ib "%i" (fun i -> i) in
  assert (debug "r_int" "finish reading");
  i

and r_native_int ib =
  assert (debug "r_native_int" "start reading");
  let i = Scanf.bscanf ib "%ni" (fun i -> i) in
  assert (debug "r_native_int" "finish reading");
  i

and r_int32 ib =
  assert (debug "r_int32" "start reading");
  let i = Scanf.bscanf ib "%li" (fun i -> i) in
  assert (debug "r_int32" "finish reading");
  i

and r_int64 ib =
  assert (debug "r_int64" "start reading");
  let i = Scanf.bscanf ib "%Li" (fun i -> i) in
  assert (debug "r_int64" "finish reading");
  i

and r_float ib =
  assert (debug "r_float" "start reading");
  let f = Scanf.bscanf ib " %f" (fun f -> f) in
  assert (debug "r_float" "finish reading");
  f

and r_float_bin ib =
  assert (debug "r_float_bin" "start reading");
  let f = Upervasives.float_of_bit_string (read_8_bytes ib) in
  assert (debug "r_float_bin" "finish reading");
  f
;;

let r_float_choose = function
  | Float_type -> r_float
  | Float_bin_type -> r_float_bin
  | Bool_type
  | String_type
  | Int_type | Native_int_type | Int32_type | Int64_type
  | Couple_type _ | Triple_type _ -> assert false
;;

(* r_tuple_lexem *)
let r_bool_couple = r_couple r_bool
and r_bool_triple = r_triple r_bool
and r_int_couple = r_couple r_int
and r_int_triple = r_triple r_int
and r_native_int_couple = r_couple r_native_int
and r_native_int_triple = r_triple r_native_int
and r_int32_couple = r_couple r_int32
and r_int32_triple = r_triple r_int32
and r_int64_couple = r_couple r_int64
and r_int64_triple = r_triple r_int64
and r_float_couple t = r_couple (r_float_choose t)
and r_float_triple t = r_triple (r_float_choose t)
;;

(* for scalars *)

let r_scalar t ib =
  match t with
  | Bool_type -> Sbool (r_bool ib)
  | Couple_type (Bool_type, Bool_type) ->
      Sbool_couple (r_bool_couple ib)
  | Triple_type (Bool_type, Bool_type, Bool_type) ->
      Sbool_triple (r_bool_triple ib)
  | String_type -> Sstring (r_string ib)
  | Int_type -> Sint (r_int ib)
  | Couple_type (Int_type, Int_type) -> Sint_couple (r_int_couple ib)
  | Triple_type (Int_type, Int_type, Int_type) -> Sint_triple (r_int_triple ib)
  | Native_int_type -> Snative_int (r_native_int ib)
  | Couple_type (Native_int_type, Native_int_type) ->
      Snative_int_couple (r_native_int_couple ib)
  | Triple_type (Native_int_type, Native_int_type, Native_int_type) ->
      Snative_int_triple (r_native_int_triple ib)
  | Int32_type -> Sint32 (r_int32 ib)
  | Couple_type (Int32_type, Int32_type) -> Sint32_couple (r_int32_couple ib)
  | Triple_type (Int32_type, Int32_type, Int32_type) ->
      Sint32_triple (r_int32_triple ib)
  | Int64_type -> Sint64 (r_int64 ib)
  | Couple_type (Int64_type, Int64_type) -> Sint64_couple (r_int64_couple ib)
  | Triple_type (Int64_type, Int64_type, Int64_type) ->
      Sint64_triple (r_int64_triple ib)
  | Float_type | Float_bin_type as t -> Sfloat (r_float_choose t ib)
  | Couple_type (Float_type, (Float_type as t))
  | Couple_type (Float_bin_type, (Float_bin_type as t)) ->
    Sfloat_couple (r_float_couple t ib)
  | Triple_type (Float_type, Float_type, (Float_type as t))
  | Triple_type (Float_bin_type, Float_bin_type, (Float_bin_type as t)) ->
    Sfloat_triple (r_float_triple t ib)
  | Couple_type (_, _)
  | Triple_type (_, _, _) -> fatal_error "r_scalar" "unexpected value type"
;;

(* for vectors *)

let r_item r_lexem ib =
  assert (debug "r_item" "start reading");
  let item = r_lexem ib in
  Scanf.bscanf ib ";\n" ();
  assert (debug "r_item" "finish reading");
  item
;;

let r_items n r_lexem ib =
  if n = 0 then [||] else
  let v0 = r_item r_lexem ib in
  let v = Array.make n v0 in
  for i = 1 to n - 1 do
    v.(i) <- r_item r_lexem ib
  done;
  v
;;

let r_vector r_lexem ib =
  assert (debug "r_vector" "start reading");
  Scanf.bscanf ib "[1\n" ();
  assert (debug "r_vector" "\"[1\" marker read");
  let n = r_size ib in
  let v = r_items n r_lexem ib in
  Scanf.bscanf ib "1]" ();
  assert (debug "r_vector" "finish reading");
  v
;;

let r_vect ib = function
  | Bool_type ->
      Vbool (r_vector r_bool ib)
  | Couple_type (Bool_type, Bool_type) ->
      Vbool_couple (r_vector r_bool_couple ib)
  | Triple_type (Bool_type, Bool_type, Bool_type) ->
      Vbool_triple (r_vector r_bool_triple ib)
  | String_type ->
      Vstring (r_vector r_string ib)
  | Int_type ->
      Vint (r_vector r_int ib)
  | Couple_type (Int_type, Int_type) ->
      Vint_couple (r_vector r_int_couple ib)
  | Triple_type (Int_type, Int_type, Int_type) ->
      Vint_triple (r_vector r_int_triple ib)
  | Native_int_type ->
      Vnative_int (r_vector r_native_int ib)
  | Couple_type (Native_int_type, Native_int_type) ->
      Vnative_int_couple (r_vector r_native_int_couple ib)
  | Triple_type (Native_int_type, Native_int_type, Native_int_type) ->
      Vnative_int_triple (r_vector r_native_int_triple ib)
  | Int32_type ->
      Vint32 (r_vector r_int32 ib)
  | Couple_type (Int32_type, Int32_type) ->
      Vint32_couple (r_vector r_int32_couple ib)
  | Triple_type (Int32_type, Int32_type, Int32_type) ->
      Vint32_triple (r_vector r_int32_triple ib)
  | Int64_type ->
      Vint64 (r_vector r_int64 ib)
  | Couple_type (Int64_type, Int64_type) ->
      Vint64_couple (r_vector r_int64_couple ib)
  | Triple_type (Int64_type, Int64_type, Int64_type) ->
      Vint64_triple (r_vector r_int64_triple ib)
  | Float_type | Float_bin_type as t ->
      Vfloat (r_vector (r_float_choose t) ib)
  | Couple_type (Float_type, (Float_type as t))
  | Couple_type (Float_bin_type, (Float_bin_type as t)) ->
      Vfloat_couple (r_vector (r_float_couple t) ib)
  | Triple_type (Float_type, Float_type, (Float_type as t))
  | Triple_type (Float_bin_type, Float_bin_type, (Float_bin_type as t)) ->
      Vfloat_triple (r_vector (r_float_triple t) ib)
  | Couple_type (_, _)
  | Triple_type (_, _, _) -> fatal_error "r_vect" "unexpected value type"
;;

(* for matrices *)

let r_line n2 r_lexem ib =
  Scanf.bscanf ib "[|\n" ();
  let v = r_items n2 r_lexem ib in
  Scanf.bscanf ib "|]" ();
  v
;;

let r_line_item n2 r_lexem ib =
  let line = r_line n2 r_lexem ib in
  Scanf.bscanf ib ";\n" ();
  line
;;

let r_line_items (n1, n2) r_lexem ib =
  if n1 = 0 then [||] else
  let m0 = r_line_item n2 r_lexem ib in
  let m = Array.make n1 m0 in
  let rec loop i =
    if i >= n1 then m else
    begin
      m.(i) <- r_line_item n2 r_lexem ib;
      loop (i + 1)
    end in
  loop 1
;;

let r_matrix t (n1, n2) layout ib =
  match t with
  | Float_type | Float_bin_type ->
    let r_float = r_float_choose t in
    begin match layout with
    | C | Permutation (0, [1]) ->
        Mcfloat (Umatrix.of_array_array (r_line_items (n1, n2) r_float ib))
    | F | Permutation (1, [0]) ->
        Mffloat
          (Umatrix.transpose
             (Umatrix.of_array_array (r_line_items (n2, n1) r_float ib)))
    | Permutation _ -> failwith "Read: pD permutation not yet implemented." end
  | Bool_type
  | String_type
  | Int_type | Native_int_type | Int32_type | Int64_type
  | Couple_type _ | Triple_type _ ->
      fatal_error "r_matrix" "unexpected value type"
;;

let r_mat ib t =
  assert (debug "r_mat" "start reading");
  Scanf.bscanf ib "[2\n" ();
  assert (debug "r_mat" "\"[2\" marker read");
  let dim = r_size_matrix ib in
  let layout = r_layout ib in
  Layout.check 2 layout;
  let m = r_matrix t dim layout ib in
  Scanf.bscanf ib "2]" ();
  assert (debug "r_mat" "finish reading");
  m
;;

let r_array ib = function
  | Vector_type t -> Vector (r_vect ib t)
  | Matrix_type t -> Matrix (r_mat ib t)
  | Scalar_type _ -> assert false
;;

let r_value ib = function
  | Scalar_type t -> Scalar (r_item (r_scalar t) ib)
  | Vector_type _ | Matrix_type _ as t ->
    let a = r_array ib t in
    Scanf.bscanf ib ";\n" ();
    a
;;

let read_typed_value ib =
  assert (debug "read_typed_value" "start reading");
  Scanf.bscanf ib "begin\n" ();
  assert (debug "read_typed_value" "\"begin\" marker read");
  let t = r_value_type ib in
  Scanf.bscanf ib "\n" ();
  assert (debug "read_typed_value" "type of value read");
  let v = r_value ib t in
  assert (debug "read_typed_value" "value read");
  Scanf.bscanf ib "end\n\n" ();
  assert (debug "read_typed_value" "\"end\" marker read");
  assert (debug "read_typed_value" "finish reading");
  v
;;
