(***********************************************************************)
(*                                                                     *)
(*                                Sklml                                *)
(*                                                                     *)
(*                 The OCaml parallel skeleton system                  *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2007-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Quentin Carbonneaux <Quentin.Carbonneaux@inria.fr>                 *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(*  This example is based on an example designed for OCamlP3l.         *)
(*                                                                     *)
(***********************************************************************)

(*****************************************************************************)
(* Program solving Poisson's equation on a rectangle with homogeneous        *)
(* Dirichlet boundary conditions, ie                                         *)
(*          { - Laplacian u = f  on Omega,                                   *)
(*          {             u = 0  on Gamma = boundary of Omega.               *)
(* The Partial Derivative Equation (PDE) is solved using Domain              *)
(* Decomposition with overlapping: firstly, on each subdomain, we input some *)
(* specific Dirichlet conditions, solve the PDE and then output Dirichlet    *)
(* conditions for the neighbor subdomains; secondly, the solution is         *)
(* searched for as the fix point of the previously defined Dirichlet to      *)
(* Dirichlet operator. Computation on each subdomain can be run in parallel. *)
(*                                                                           *)
(* The CamlP3l implementation uses the loop skeleton for the fix point       *)
(* and the mapvector skeleton for the concurrent computation on each         *)
(* subdomain (see the Make_domaindecomposition module).                      *)
(* The poisson's solver is a C program implementing an implicit 5-point      *)
(* scheme which is spawned and connected via stdin/stdout to                 *)
(* computation_on_a_subdomain in a seq skeleton (see "poisson.c").           *)
(* Graphical facilities using gnuplot are provided (see Gnuplot module).     *)
(*                                                                           *)
(* The Rdec2 module provides functions to manipulate regular decompositions  *)
(* of regular rectangular grids.                                             *)
(*****************************************************************************)

(* $Id: main.ml,v 1.26 2014/01/21 09:21:18 fclement Exp $ *)

open Skl;;

(* The atomic skeletons. *)

let plot_skl =
  skl (config, rdec) -> Make_dom2.plot config rdec
;;

let subdomain_worker_skl =
  skl (config, sd) -> Make_dom2.subdomain_worker config sd
;;

let convergence_skl =
  skl config -> Make_dom2.convergence config
;;

(* The full computational skeleton expression. *)
let domain_skl (cfg_fname, cfg, rdec) =
  let nw = Options.get_number_of_workers () in
  Printf.eprintf "cfg_fname = %s\nnworkers = %d\n" cfg_fname nw;
  let mk_worker i =
    (subdomain_worker_skl (cfg, (rdec, i)),
     i :: Rdec2.neighbors_of (rdec, i)) in
  let spec = Array.init nw mk_worker in
  Make_domain.make_domain_opt spec
    (plot_skl (cfg, rdec) ||| convergence_skl cfg)
;;

let generate_input_stream rdec =
  Sklstream.singleton (Make_dom2.set_data rdec)
;;

let finalize cfg cfg_fname = Make_dom2.print_result cfg cfg_fname;;

let main { Parfun.parfun = parfun; } =
  Ucommand_option.parse ();
  let (cfg_fname, cfg, rdec) as gp = Options.get_global_parameters () in
  let input_stream = generate_input_stream rdec in
  let output_stream = parfun (domain_skl gp) input_stream in
  Sklstream.iter ignore output_stream;
  finalize cfg cfg_fname;
;;

Options.add_options ();;
Parfun.pardo main;;
