(***********************************************************************)
(*                                                                     *)
(*                                Sklml                                *)
(*                                                                     *)
(*                 The OCaml parallel skeleton system                  *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2007-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Quentin Carbonneaux <Quentin.Carbonneaux@inria.fr>                 *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: parser_char.ml,v 1.5 2014/01/21 09:21:18 fclement Exp $ *)

open Parser;;

let string_skipn n s =
  String.sub s n (String.length s - n)
;;

let implode l =
  let rec implode i l =
    match l with
    | [] -> String.create i
    | c :: t ->
      let s = implode (succ i) t in
      s.[i] <- c; s in
  implode 0 l
;;

let pchar c =
  { parse = fun s ->
    if String.length s = 0 then None else
    if s.[0] = c then Some (s.[0], string_skipn 1 s) else
    None }
;;

let pchar_ign c = liftM ignore (pchar c);;

let pclass l = any (List.map pchar l);;

let pnchar c =
  { parse = fun s ->
    if String.length s = 0 then None else
    if s.[0] <> c then Some (s.[0], string_skipn 1 s) else
    None }
;;

let pword w =
  let l = String.length w in
  let rec pword i =
    if i >= l then return () else
    pchar w.[i] >> pword (succ i) in
  pword 0
;;

let pspaces = liftM ignore (pclass [ ' '; '\r'; '\n'; '\t'; ]);;

let pspaced p = many pspaces >> p;;

let pint =
  let pnum1 =
    let digits = [ '0'; '1'; '2'; '3'; '4'; '5'; '6'; '7'; '8'; '9'; ] in
    let int_of_char c = Char.code c - Char.code '0' in
    liftM int_of_char (pclass digits) in
  let rec pint n =
    (pnum1 >>=
      fun d -> pint (n * 10 + d)) +++
    (return n) in
  maybe (pchar_ign '-') >>= fun signo ->
    match signo with
    | Some () -> liftM (fun x -> -x) (pnum1 >>= pint)
    | None -> pnum1 >>= pint
;;

let pfloat =
  let pnum1 =
    let digits = [ '0'; '1'; '2'; '3'; '4'; '5'; '6'; '7'; '8'; '9'; ] in
    let int_of_char c = Char.code c - Char.code '0' in
    liftM int_of_char (pclass digits) in
  let rec pdecimal n =
    (pnum1 >>=
      fun d -> pdecimal (n *. 10. +. float_of_int d)) +++
    (pchar_ign '.' >>
      (pmantissa 10. 0. >>= fun s -> return (n +. s))) +++
    (return n)
  and pmantissa k s =
    (pnum1 >>=
      fun d -> pmantissa (k *. 10.) (float_of_int d /. k +. s)) +++
    (return s) in
  maybe (pchar_ign '-') >>= fun signo ->
    match signo with
    | Some () ->
      liftM (fun x -> -. x) (liftM float_of_int pnum1 >>= pdecimal)
    | None -> liftM float_of_int pnum1 >>= pdecimal
;;
