(***********************************************************************)
(*                                                                     *)
(*                                Sklml                                *)
(*                                                                     *)
(*                 The OCaml parallel skeleton system                  *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2007-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Quentin Carbonneaux <Quentin.Carbonneaux@inria.fr>                 *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: tridiag.ml,v 1.5 2014/01/21 09:21:18 fclement Exp $ *)

type t = (float array * float array * float array);;

type pvect =
  | Top of float array * float               (* elmts, bot *)
  | Mid of int * float * float array * float (* index, top, elmts, bot *)
  | Bot of float * float array               (* top. elmts *)
;;

let create diag0 diag1 diag2 =
  if Array.length diag0 <> Array.length diag2 ||
     Array.length diag0 <> Array.length diag1 - 1 then
    invalid_arg "Tridiag.create" else
  (diag0, diag1, diag2)
;;

let pmult (d0, d1, d2) pvect =
  let n = Array.length d1 in
  let compute_middle f arr elmts start stop =
    for i = start to stop do
      arr.(i) <-
        elmts.(i - 1) *. d0.(f (i - 1)) +.
        elmts.(i) *. d1.(f i) +.
        elmts.(i + 1) *. d2.(f i)
    done in
  match pvect with
  | Top (elmts, bot) ->
    let nelmts = Array.length elmts in
    let out = Array.create nelmts 0. in
    compute_middle (fun x -> x) out elmts 1 (nelmts - 2);
    out.(0) <- elmts.(0) *. d1.(0) +. elmts.(1) *. d2.(0);
    out.(nelmts - 1) <-
      elmts.(nelmts - 2) *. d0.(nelmts - 2) +.
      elmts.(nelmts - 1) *. d1.(nelmts - 1) +.
      bot *. d2.(nelmts - 1);
    out
  | Bot (top, elmts) ->
    let nelmts = Array.length elmts in
    let out = Array.create nelmts 0. in
    compute_middle (fun x -> n - nelmts + x) out elmts 1 (nelmts - 2);
    out.(0) <-
      top *. d0.(n - nelmts - 1) +.
      elmts.(0) *. d1.(n - nelmts) +.
      elmts.(1) *. d2.(n - nelmts);
    out.(nelmts - 1) <-
      elmts.(nelmts - 2) *. d0.(n - 2) +. elmts.(nelmts - 1) *. d1.(n - 1);
    out
  | Mid (idx, top, elmts, bot) ->
    let nelmts = Array.length elmts in
    let out = Array.create nelmts 0. in
    compute_middle (fun x -> idx + x) out elmts 1 (nelmts - 2);
    out.(0) <-
      top *. d0.(idx - 1) +.
      elmts.(0) *. d1.(idx) +.
      elmts.(1) *. d2.(idx);
    out.(nelmts - 1) <-
      elmts.(nelmts - 2) *. d0.(idx + nelmts - 2) +.
      elmts.(nelmts - 1) *. d1.(idx + nelmts - 1) +.
      bot *. d2.(idx + nelmts - 1);
    out
;;

let mk_pvect a =
  let chk b =
    if not b then invalid_arg "Tridiag.mk_pvect" else () in
  match Array.length a with
  | 0 -> invalid_arg "Tridiag.mk_pvect"
  | 1 ->
    let a = a.(0) in
    let l = Array.length a in
    chk (l > 1);
    [| Mid (1, a.(0), Array.sub a 1 (l - 2), a.(l - 1)); |]
  | 2 ->
    let (a0, a1) = (a.(0), a.(1)) in
    let (l0, l1) = (Array.length a0, Array.length a1) in
    chk (l0 > 0 && l1 > 0);
    [| Top (a0, a1.(0)); Bot (a0.(l0 - 1), a1); |]
  | l ->
    let (a0, _al) = (a.(0), a.(l - 1)) in
    let l2 = Array.length a.(l - 2) in
    let out = Array.create l (Bot (0., [||])) in
    let idx = ref (Array.length a0) in
    for i = 1 to l - 2 do
      let li = Array.length a.(i - 1) in
      chk (li > 0);
      out.(i) <-
        Mid (!idx, a.(i - 1).(li - 1), a.(i), a.(i + 1).(0));
      idx := !idx + Array.length a.(i)
    done;
    out.(0) <- Top (a.(0), a.(1).(0));
    out.(l - 1) <- Bot (a.(l - 2).(l2 - 1), a.(l - 1));
    out
;;

let split_vect n v =
  if n = 0 then invalid_arg "Tridiag.split_vect" else
  let l = Array.length v / n in
  if l = 0 then invalid_arg "Tridiag.split_vect" else
  let rec loop n i vecs =
    if n = 1
      then Array.of_list (Array.sub v i (Array.length v - i) :: vecs) else
    let vecs = Array.sub v i l :: vecs in
    loop (pred n) (i + l) vecs in
  loop n 0 []
;;
