(***********************************************************************)
(*                                                                     *)
(*                                Sklml                                *)
(*                                                                     *)
(*                 The OCaml parallel skeleton system                  *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2007-2012 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Quentin Carbonneaux <Quentin.Carbonneaux@inria.fr>                 *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: options.ml,v 1.7 2013/03/29 13:21:41 fclement Exp $ *)

let min_number_of_workers = 1
and min_pool_size = 20
and min_number_of_factors = 4
and min_factor_size = 256
;;

let default_number_of_workers = 2
and default_pool_size = min_pool_size
and default_number_of_factors = min_number_of_factors
and default_factor_size = min_factor_size
;;

let my_make_aref with_assignment name default n =
  Uaref.make_aref_gen
    (Uaref.with_int_bounds (Some n) None
      (Uaref.with_name name
        (with_assignment
          (Uaref.aref_spec default))))
;;

let make_single_aref_with_min name =
  my_make_aref Uaref.with_single_assignment name
;;

let { Uaref.get = get_number_of_workers; set = set_number_of_workers; } =
  make_single_aref_with_min
    "number_of_workers" default_number_of_workers min_number_of_workers

and { Uaref.get = get_pool_size; set = set_pool_size; } =
  make_single_aref_with_min
    "pool_size" default_pool_size min_pool_size

and { Uaref.get = get_number_of_factors; set = set_number_of_factors; } =
  make_single_aref_with_min
    "number_of_factors" default_number_of_factors min_number_of_factors

and { Uaref.get = get_factor_size; set = set_factor_size; } =
  make_single_aref_with_min
    "factor_size" default_factor_size min_factor_size

and {Uaref.get = get_pool_directory; set = set_pool_directory; } =
  Uaref.make_aref_single
    "pool_directory "
    (Filename.concat (Sys.getcwd ()) "work")

and {Uaref.get = get_temporary_directory; set = set_temporary_directory; } =
  Uaref.make_aref_single
    "temporary_directory "
    (Filename.concat (Sys.getcwd ()) (Filename.concat "work" "tmp"))

and {Uaref.get = get_cbits_directory; set = set_cbits_directory; } =
  Uaref.make_aref_single
    "cbits_directory"
    (Filename.concat (Sys.getcwd ()) "cbits")
;;

let get_prime_gen_path () =
  Filename.concat (get_cbits_directory ()) "prime_gen"
;;

let get_prime_check_path () =
  Filename.concat (get_cbits_directory ()) "prime_check"
;;

exception Parse_failure of string;;

let parse_recover_data s =
  let ib = Scanf.Scanning.from_string s in
  let check_id mx x =
    if x < 0 || x >= mx || x >= get_pool_size ()
    then raise (Parse_failure "invalid permutation")
    else x
  in
  let rec loop mx perm =
    if Scanf.Scanning.end_of_input ib then List.rev perm else
    let x =
      if Scanf.bscanf ib "%0c" (fun c -> c = ':')
      then Scanf.bscanf ib "%_c%d" (check_id mx)
      else Scanf.bscanf ib "%d" (check_id mx)
    in
    loop x (x :: perm)
  in
  try
    let perm = loop max_int [] in
    if List.length perm <> get_number_of_factors ()
    then raise (Parse_failure "incorrect number of ids")
    else Some perm
  with
  | Scanf.Scan_failure s ->
      raise (Parse_failure ("scan failure - " ^ s))
  | End_of_file ->
      raise (Parse_failure "premature end of input")
;;

let get_recover_data, set_recover_data, unsafe_set_recover_data =
  let { Uaref.get = get; set = set; } =
    Uaref.make_aref_single "recover_data" None in
  (get, (fun s -> set (parse_recover_data s)), set)
;;

type configuration = {
  number_of_workers : int;
  pool_size : int;
  number_of_factors : int;
  factor_size : int;
  pool_directory : string;
  temporary_directory : string;
  cbits_directory : string;
  recover_data : (int list) option
};;

let save_config () = {
  number_of_workers = get_number_of_workers ();
  pool_size = get_pool_size ();
  number_of_factors = get_number_of_factors ();
  factor_size = get_factor_size ();
  pool_directory = get_pool_directory ();
  temporary_directory = get_temporary_directory ();
  cbits_directory = get_cbits_directory ();
  recover_data = get_recover_data ();
};;

let set_config c =
  try
    set_number_of_workers c.number_of_workers;
    set_pool_size c.pool_size;
    set_number_of_factors c.number_of_factors;
    set_factor_size c.factor_size;
    set_pool_directory c.pool_directory;
    set_temporary_directory c.temporary_directory;
    set_cbits_directory c.cbits_directory;
    unsafe_set_recover_data c.recover_data
  with
    Failure _ -> () (* we are in sequential mode ... *)
;;

let add_options () =
  let check_dir_and_set f d =
    if
      try Sys.is_directory d with _ -> false
    then f d
    else failwith (Printf.sprintf "%s is not a valid directory" d)
  in
  
  Ucommand_option.user_add_all
    (List.map
       (fun (rco, sk_opt) ->
         let with_short_key =
           match sk_opt with
           | None -> Ucommand_option.with_auto_short_key
           | Some sk -> Ucommand_option.with_short_key sk in
         with_short_key (Ucommand_option.of_regular rco))
       [
        ({
          Ucommand_option.key = "--number-of-worker";
          spec = Arg.Int set_number_of_workers;
          doc =
            Printf.sprintf
              "<int>  Set the number of workers.\
              \n\tDefault is %d."
              (get_number_of_workers ());
         },
         Some "-nw");
        
        ({
          Ucommand_option.key = "--pool-size";
          spec = Arg.Int set_pool_size;
          doc =
            Printf.sprintf
              "<int>  Set the size of the prime pool.\
              \n\tDefault is %d.\
              \n\tCan't be lower than %d."
              (get_pool_size ()) min_pool_size;
         },
         None);
        
        ({
          Ucommand_option.key = "--number-of-factors";
          spec = Arg.Int set_number_of_factors;
          doc =
            Printf.sprintf
              "<int>  Set the number of big factors in p-1.\
              \n\tDefault is %d."
              (get_number_of_factors ());
         },
         Some "-nf");
        
        ({
          Ucommand_option.key = "--factor-size";
          spec = Arg.Int set_factor_size;
          doc =
            Printf.sprintf
              "<int>  Set the size of factors in bits.\
              \n\tDefault is %d.\
              \n\tCan't be lower than %d."
              (get_factor_size ()) min_factor_size;
         },
         None);
        
        ({
          Ucommand_option.key = "--pool-directory";
          spec = Arg.String (check_dir_and_set set_pool_directory);
          doc =
            Printf.sprintf
              "<dir>  Set the exchange directory (used to serialize data).\
              \n\tDefault is %S."
              (get_pool_directory ());
         },
         None);
        
        ({
          Ucommand_option.key = "--temporary-directory";
          spec = Arg.String (check_dir_and_set set_temporary_directory);
          doc =
            Printf.sprintf
              "<dir>  Set the temporary directory used during computation.\
              \n\tDefault is %S."
              (get_temporary_directory ());
         },
         None);
        
        ({
          Ucommand_option.key = "--cbits-directory";
          spec = Arg.String (check_dir_and_set set_cbits_directory);
          doc =
            Printf.sprintf
              "<dir>  Set the directory containing prime_gen prime_check.\
              \n\tDefault is %S."
              (get_cbits_directory ());
         },
         None);
        
        ({
          Ucommand_option.key = "--recover";
          spec = Arg.String set_recover_data;
          doc =
            "<string>  Warm restart from the given permutation.\
            \n\tPermutation is in the following format x1:x2:x3:..:xi where\
            \n\t + pool_size > x1 > x2 > ... > xi >= 0\
            \n\t + i is number_of_factors.";
         },
         Some "-re");
       ]);
;;

Ucommand_option.set_usage_msg
  (Printf.sprintf "\n%s -- sklml demo code\n" Sys.argv.(0))
;;
