(***********************************************************************)
(*                                                                     *)
(*                                Sklml                                *)
(*                                                                     *)
(*                 The OCaml parallel skeleton system                  *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2007-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Quentin Carbonneaux <Quentin.Carbonneaux@inria.fr>                 *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: darray.ml,v 1.11 2014/01/21 09:21:15 fclement Exp $ *)

(* Dynamic arrays, inserting elements on the fly is made easy,
  retreiving is in O(1) *)

exception Out_of_memory

type increase_mode_t =
    | Add of int
    | Double

module type CONTROL =
  sig
    val min_size : int
    val increase_mode : increase_mode_t
  end

module type S =
  sig
    type 'a t
    val create : int -> 'a t
    val length : 'a t -> int
    val append : 'a t -> 'a -> int
    val get : 'a t -> int -> 'a
    val set : 'a t -> int -> 'a -> unit
    val fold_left : ('a -> 'b -> 'a) -> 'a -> 'b t -> 'a
  end

module Make(C : CONTROL) =
  struct
    (* size, number set, data*)
    type 'a t = (int * int * 'a option array) ref

    let increase =
      match C.increase_mode with
      | Double ->
        fun i -> if i = 0 then C.min_size else 2 * i
      | Add j -> fun i -> i + j

    let realloc a =
      let s, x, arr = !a in
      if x < s then () else
      let nsize = increase s in
      if nsize > Sys.max_array_length
        then raise Out_of_memory else
      let narray = Array.make nsize None in
      Array.blit arr 0 narray 0 s;
      a := nsize, x, narray

    let create i =
      if i < 1
        then invalid_arg "Darray.Make().create: size negative" else
      let i = max i C.min_size in
      ref (i, 0, Array.make i None)

    let length a =
      let _, n, _ = !a in n

    let append a x =
      realloc a;
      let s, n, arr = !a in
      arr.(n) <- Some x;
      a := s, n + 1, arr;
      n

    let get a i =
      let _, n, arr = !a in
      if i >= n
        then invalid_arg "Darray.Make().get: index too big" else
      match arr.(i) with Some x -> x | None -> assert false

    let set a i x =
      let _, n, arr = !a in
      if i >= n
      then invalid_arg "Darray.Make().set: index too big" else
      arr.(i) <- Some x

    let fold_left f z a =
      let _, _, arr = !a in
      let ff a b =
        match b with
        | Some x -> f a x
        | None -> a in
      Array.fold_left ff z arr
  end

module Base_control : CONTROL =
  struct
    let increase_mode = Double
    let min_size = 16
  end

module A = Make(Base_control)
