(***********************************************************************)
(*                                                                     *)
(*                                Sklml                                *)
(*                                                                     *)
(*                 The OCaml parallel skeleton system                  *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2007-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Quentin Carbonneaux <Quentin.Carbonneaux@inria.fr>                 *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: main.ml,v 1.14 2014/01/21 09:21:16 fclement Exp $ *)

open Sklml_install_data;;

type dependency_options = {
  lib_root_dir : string;
  pp_cmd : string;
  ocamldep_options : string list;
};;

let unwords = String.concat " ";;

let mk_command {
    lib_root_dir = _;
    pp_cmd = pp_cmd;
    ocamldep_options = camldep_options;
  } =
  let pp_option = [
    "-pp";
    Printf.sprintf "'%s'" pp_cmd;
  ] in
  unwords ([ "ocamldep"; ] @ pp_option @ camldep_options)
;;

let fail msg =
  Printf.eprintf "sklmldep: %s\n" msg; exit 1
;;

let parse_argv args =
  let (get_annons, accum_annons) =
      let annons = ref [] in
      (fun () -> !annons),
      (fun px s -> annons := !annons @ [px ^ s])
  and (get_lib_root_dir, set_lib_root_dir) =
    let dir = ref default_lib_root_dir in
    (fun () -> !dir),
    (fun s -> dir := s)
(*  and (get_pp, set_pp) =
    let pp = ref camlp45_pa_skl in
    (fun () -> !pp),
    (fun s -> pp := s)*)
  and (get_debug, set_debug) =
    let d = ref false in
    (fun () -> !d),
    (fun () -> d := true)
  in
  let speclist =
    [
     "-sklml-lib-root-dir", Arg.String set_lib_root_dir,
     "<dir>  Set the library root directory of the Sklml system.";
      
(*     "-pp", Arg.String set_pp,
     "<pp>  Set the OCaml pre-processor.";*)
     
     "-ext", Arg.String (accum_annons "-"),
     "<arg>  Internal use only.";
     
     "-debug", Arg.Unit set_debug,
     " Turn the debugging mode on. Default is off.";
     
     "-where",
     Arg.Unit (fun () -> prerr_endline (get_lib_root_dir ()); exit 0),
     " Print the location of Sklml system and exit.";
     
     "-v",
     Arg.Unit
       (Upph_id_card.make_print_short_version Sklml_version.pph_id_card),
     " Print the short form of the version information and exit.";
     
     "-version",
     Arg.Unit
       (Upph_id_card.make_print_long_version Sklml_version.pph_id_card),
     " Print the long form of the version information and exit.";
    ] in
  let keys =
    [ "-help"; "--help"; ] @ List.map (fun (x, _, _) -> x) speclist in
  let args = Array.of_list
    (Array.fold_left (fun l s ->
      if not (List.mem s keys) then
        let key = String.sub s 1 (String.length s - 1) in
        if s.[0] = '-'
        then l @ [ "-ext"; key ]
        else l @ [s]
      else l @ [s]) [] args) in
  let head_usage =
    "Usage: sklmldep [OPTIONS] FILES...\n\n\
     Options are (in addition to those of the OCaml dependency generator):" in
  try
    Arg.parse_argv args speclist (accum_annons "") head_usage;
    let lib_root_dir = get_lib_root_dir () in
    ({
       lib_root_dir = lib_root_dir;
       pp_cmd = mk_pp_cmd (mk_camlp45_pa_skl lib_root_dir);
       ocamldep_options = get_annons ();
     }, get_debug ())
  with
  | Arg.Bad _ -> fail "invalid usage, see help using -help option"
  | Arg.Help m -> print_string m; exit 0
  | Failure m -> fail ("invalid usage, " ^ m)
;;

let dprintf b =
  (if b then Printf.fprintf else Printf.ifprintf) stdout
;;

let main () =
  let (ocamldep_opts, dbg) = parse_argv Sys.argv in
  let compile_cmd = mk_command ocamldep_opts in
  dprintf dbg "will run: %s\n" compile_cmd;
  let ret = Sys.command compile_cmd in
  dprintf dbg
    (if ret = 0
     then format_of_string "dependencies generated\n"
     else format_of_string "there were errors\n");
  exit ret
;;

let _ = main ();;
