(***********************************************************************)
(*                                                                     *)
(*                                Sklml                                *)
(*                                                                     *)
(*                 The OCaml parallel skeleton system                  *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2007-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Quentin Carbonneaux <Quentin.Carbonneaux@inria.fr>                 *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: master.ml,v 1.33 2014/01/21 09:21:16 fclement Exp $ *)

type node = {
   local_addr : Unix.inet_addr;
   nslaves : int;
   skl_port : int;
   slave_port : int;
};;

type slave = {
   io : IO.ic * IO.oc;
   addr : Unix.inet_addr;
   col : int;
};;

type waiting_state = {
   slaves_seen : slave list;
   slave_count : int;
   listen_sock : Unix.file_descr;
};;

type serving_state = {
   slaves : slave list;
   sklml_chans : IO.ic * IO.oc;
};;

type state =
   | Start
   | Waiting_slaves of waiting_state
   | Start_serving of slave list
   | Serving of serving_state
   | Stop
;;

type request =
   | Slave_port_available of Unix.file_descr
   | Slave_connect of slave * waiting_state
   | Client_available of serving_state
   | Client_request of Skl_comm.request * serving_state
   | Client_lost
;;

type placement_policy =
   | First_fit (* Choose the first fitting bin *)
   | Swap (* Used when the job can't be assigned (colors become negative) *)
   | Worst_fit (* Choose the emptiest bin *)
;;
(* Placement policies used to assign a job to a processor, here, they
  are sorted by speed. *)

(* val update : ('a -> 'a) -> 'a -> 'a list -> 'a list *)
let update f a l =
  let rec update acc l =
    match l with
    | [] -> invalid_arg "Master.update"
    | h :: t ->
      if h == a
      then List.rev_append acc (f h :: t)
      else update (h :: acc) t in
  update [] l
;;
(* [update f a l] updates the field [a] of [l] using the function
  [f]. Like [map], on a single element. *)

(* val extract : 'a -> 'a list -> 'a list *)
let extract a l =
  let rec extract acc l =
    match l with
    | [] -> invalid_arg "Master.extract"
    | h :: t ->
      if h == a
      then List.rev_append acc t
      else extract (h :: acc) t in
  extract [] l
;;
(* [extract a l] updates the list [l] by removing the element [a]. If
  this one is not in the list it fails with an [Invalid_arg]
  exception. *)

(* val max : ('a -> 'a -> bool) -> 'a list -> 'a *)
let max cmp l =
  let rec max m l =
    match l with
    | [] -> m
    | h :: t ->
      max (if cmp h m then h else m) t in
  match l with
  | [] -> invalid_arg "Master.max"
  | h :: t -> max h t
;;
(* [max cmp l] returns the biggest element of the list [l] using the
  order [cmp]. [cmp a b] is true iff '[a] > [b]'. *)

(* val placement : placement_policy -> slave list -> int ->
  slave * slave list *)
let rec placement policy slaves job =
  match policy with
  | First_fit ->
    let rec loop acc l =
      match l with
      | ({ col = c; _ } as slave) :: t ->
        if c >= job
        then let s = { slave with col = (c - job); } in
          (slave, List.rev_append acc (s :: t))
        else loop (slave :: acc) t
      | [] -> (* we failed *)
        placement Swap slaves job in
    loop [] slaves
  | Worst_fit | Swap ->
    let compare a b = a.col > b.col
    and update_slave ({ col = c; _ } as s) =
      { s with col = (c - job); } in
    let slave = max compare slaves in
    let slaves = update update_slave slave slaves in
    (slave, slaves)
;;

(* Request generator *)
let req_generator node s =
  try
    match s with
    | Start ->
      let sock, addr =
        IO.create_bound_socket (Some node.slave_port) in
      Log.say Log.Say "listenning for slaves on %a"
        IO.dump_sockaddr addr;
      Some
        (Slave_port_available sock)
    | Waiting_slaves state ->
      let ((ic, _) as io), addr =
        IO.wait_connection state.listen_sock false in
      let addr =
        if addr = Unix.inet_addr_loopback
        then node.local_addr
        else addr in
      Log.say Log.Say
        "slave accepted (%s)" (Unix.string_of_inet_addr addr);
      let { Daemon_types.color = col; } = IO.of_channel ic in
      Some
        (Slave_connect
          ({ col = col; io = io; addr = addr; }, state))
    | Start_serving slaves ->
      let sock, addr =
        IO.create_bound_socket (Some node.skl_port) in
      Log.say Log.Say
        "listenning for sklml program on %a" IO.dump_sockaddr addr;
      let io, _ = IO.wait_connection sock true in
      Some
        (Client_available { sklml_chans = io; slaves = slaves; })
    | Serving ({ sklml_chans = (ic, _); _ } as state) ->
      let request =
        match
          try Some (Skl_comm.extract_request ic)
          with End_of_file -> None
        with
        | Some r ->
          Client_request (r, state)
        | None -> Client_lost in
      Some request
    | Stop -> None
  with IO.IOError msg ->
    Log.say Log.Err "IO error: %s" msg;
    None
;;

(* Request handler *)
let req_handler node req =
  match req with
  | Slave_port_available fd ->
    let state =
      { slaves_seen = []
      ; slave_count = 0
      ; listen_sock = fd
      } in
    Waiting_slaves state
  | Slave_connect (slave, waiting_state) ->
    if waiting_state.slave_count + 1 = node.nslaves then begin
      Unix.close waiting_state.listen_sock;
      Start_serving (slave :: waiting_state.slaves_seen)
    end else
      Waiting_slaves
        { waiting_state with
          slaves_seen = slave :: waiting_state.slaves_seen;
          slave_count = 1 + waiting_state.slave_count; }
  | Client_available serving_state ->
    Serving serving_state
  | Client_request (req, serving_state) ->
    let { slaves = slaves; sklml_chans = (_, oc); } =
      serving_state in
    (* maybe_rep is
      Some (daemon_rep, slave_addr, new_slaves, slaves) if the
      request was handled by a slave, and
      None if no more slaves are available *)
    (* slaves must be returned in case of a slave loss followed
      by a Spawn_failure reply. new_slaves is slaves updated (the
      color of the chosen slave is changed). slave_addr is the
      address of the chosen slave. daemon_rep is the reply of the
      slave. *)
    let maybe_rep =
      let rec send_request slaves =
        if slaves = [] then None else (* No slaves available, fail *)
        let daemon_req, slave, new_slaves =
          match req with
          | Skl_comm.Pipe_req (t1, t2) ->
            let slave, _ = placement Worst_fit slaves 0 in
            (Daemon_types.Spawn_pipe (t1, t2)), slave, slaves
          | Skl_comm.Loop_req (t1, t2) ->
            let slave, _ = placement Worst_fit slaves 0 in
            (Daemon_types.Spawn_loop (t1, t2)), slave, slaves
          | Skl_comm.Atom_req (p, atom_id, col) ->
            let slave, slaves = placement First_fit slaves col in
            (Daemon_types.Spawn_atom (p, atom_id)), slave, slaves in
        match
          try Some (Spawnlib.request_slave slave.io daemon_req)
          with End_of_file | IO.IOError _ -> None
        with
        | Some rep -> Some (rep, slave.addr, new_slaves, slaves)
        | None ->
          Log.say Log.Warn
            "slave lost, continuing without him (ip:%s)"
            (Unix.string_of_inet_addr slave.addr);
          send_request (extract slave slaves) in
      send_request slaves in
    Spawnlib.maybe maybe_rep
      (* error *)
      (fun () ->
        Log.say Log.Err "no more slaves remaining, exiting";
        Stop)
      (* success *)
      (fun (daemon_rep, slave_addr, new_slaves, slaves) ->
        let skl_reply, slaves =
          match daemon_rep with
          | Daemon_types.Spawn_success
            { Daemon_types.worker_port = p; } ->
              Skl_comm.Success (Unix.ADDR_INET (slave_addr, p)), new_slaves
          | Daemon_types.Spawn_failure ->
            Skl_comm.Failure, slaves in
        Skl_comm.send_reply oc skl_reply;
        Serving { serving_state with slaves = slaves; })
  | Client_lost -> Stop
;;

(* val master_loop : node -> state -> state *)
let master_loop node =
  Spawnlib.master_loop (req_generator node) (req_handler node)
;;

let start node =
  match master_loop node Start with
  | Stop -> ()
  | Start ->
    Spawnlib.fail "System failure in `Start' state."
  | Waiting_slaves _ ->
    Spawnlib.fail "System failure in `Waiting_slaves' state."
  | Start_serving _ ->
    Spawnlib.fail "System failure in `Start_serving' state."
  | Serving _ ->
    Spawnlib.fail "System failure in `Serving' state."
;;
